/*
 * Copyright (C) 2015 Freescale Semiconductor, Inc.
 * Copyright (C) 2016 Atmark Techno, Inc.
 *
 * SPDX-License-Identifier:	GPL-2.0+
 */

#include <asm/arch/clock.h>
#include <asm/arch/imx-regs.h>
#include <asm/arch/mx7-pins.h>
#include <asm/arch/sys_proto.h>
#include <asm/gpio.h>
#include <asm/imx-common/boot_mode.h>
#include <asm/imx-common/iomux-v3.h>
#include <asm/io.h>
#include <linux/sizes.h>
#include <common.h>
#include <fsl_esdhc.h>
#include <mmc.h>
#include <miiphy.h>
#include <netdev.h>
#include "../../freescale/common/pfuze.h"
#include <i2c.h>
#include <asm/imx-common/mxc_i2c.h>
#include <asm/arch/crm_regs.h>
#include <usb.h>
#include <usb/ehci-ci.h>
#include <wait_bit.h>
#include <watchdog.h>
#include <environment.h>

DECLARE_GLOBAL_DATA_PTR;

#define UART_PAD_CTRL  (PAD_CTL_DSE_3P3V_49OHM | \
	PAD_CTL_PUS_PU100KOHM | PAD_CTL_HYS)

#define USDHC_PAD_CTRL (PAD_CTL_DSE_3P3V_32OHM | PAD_CTL_SRE_SLOW | \
	PAD_CTL_HYS | PAD_CTL_PUE | PAD_CTL_PUS_PU47KOHM)
#define USDHC_PAD_CTRL_CD (PAD_CTL_DSE_3P3V_32OHM | PAD_CTL_SRE_SLOW | \
	PAD_CTL_HYS)
#define USDHC_PAD_CTRL_WP (PAD_CTL_DSE_3P3V_32OHM | PAD_CTL_SRE_SLOW | \
	PAD_CTL_HYS)

#define ENET_PAD_CTRL  (PAD_CTL_PUS_PU100KOHM | PAD_CTL_DSE_3P3V_49OHM)
#define ENET_PAD_CTRL_MII  (PAD_CTL_DSE_3P3V_32OHM)
#define ENET_PAD_CTRL_MII_G3L  (PAD_CTL_DSE_3P3V_32OHM | PAD_CTL_PUS_PU47KOHM)

#define ENET_RX_PAD_CTRL  (PAD_CTL_PUS_PU100KOHM | PAD_CTL_DSE_3P3V_49OHM)

#define I2C_PAD_CTRL    0x0

#define LCD_PAD_CTRL    (PAD_CTL_HYS | PAD_CTL_PUS_PU100KOHM | \
	PAD_CTL_DSE_3P3V_49OHM)

#define QSPI_PAD_CTRL	\
	(PAD_CTL_DSE_3P3V_49OHM | PAD_CTL_PUE | PAD_CTL_PUS_PU47KOHM)

#define NAND_PAD_CTRL (PAD_CTL_DSE_3P3V_49OHM | PAD_CTL_SRE_SLOW | PAD_CTL_HYS)

#define NAND_PAD_READY0_CTRL (PAD_CTL_DSE_3P3V_49OHM | PAD_CTL_PUS_PU5KOHM)

#define GPIO_PAD_CTRL (PAD_CTL_SRE_SLOW | PAD_CTL_PUE | PAD_CTL_PUS_PD100KOHM)


#ifdef CONFIG_SYS_I2C_MXC
#define PC MUX_PAD_CTRL(I2C_PAD_CTRL)

#define BOARDTYPE_AIOTG_G3  0x0A000000
#define BOARDTYPE_AIOTG_G3_D2 0x0A000001
#define BOARDTYPE_AIOTG_G3_W1 0x0A000002
#define BOARDTYPE_AIOTG_G3_M1 0x0A000003
#define BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z 0x0A000004
#define BOARDTYPE_AIOTG_G3_W2 0x0A000005
#define BOARDTYPE_AX1       0x0A100000
#define BOARDTYPE_AIOTG_G3L 0x0A200000
#define BOARDTYPE_UNKNOWN   0xFFFFFFFF

struct kotai_code {
	unsigned short product_id;
	unsigned short lot;
	unsigned short number;
};

struct board_info {
	unsigned int board_type;
	unsigned short revision;
	unsigned int serial_no;
	struct kotai_code kotai;
	unsigned char serial_no_valid;
	unsigned int dram;
	unsigned char xtal;
	unsigned char usermac[6];
	int exist_sdhc1;
};

static struct board_info x1_board_info;

/* I2C4 for PMIC */
struct i2c_pads_info i2c_pad_info4 = {
	.scl = {
		.i2c_mode = MX7D_PAD_GPIO1_IO10__I2C4_SCL | PC,
		.gpio_mode = MX7D_PAD_GPIO1_IO10__GPIO1_IO10 | PC,
		.gp = IMX_GPIO_NR(1, 10),
	},
	.sda = {
		.i2c_mode = MX7D_PAD_GPIO1_IO11__I2C4_SDA | PC,
		.gpio_mode = MX7D_PAD_GPIO1_IO11__GPIO1_IO11 | PC,
		.gp = IMX_GPIO_NR(1, 11),
	},
};

struct i2c_pads_info i2c_pad_x1_addon_info3 = {
	.scl = {
		.i2c_mode = MX7D_PAD_I2C3_SCL__I2C3_SCL | PC,
		.gpio_mode = MX7D_PAD_I2C3_SCL__GPIO4_IO12 | PC,
		.gp = IMX_GPIO_NR(4, 12),
	},
	.sda = {
		.i2c_mode = MX7D_PAD_I2C3_SDA__I2C3_SDA | PC,
		.gpio_mode = MX7D_PAD_I2C3_SDA__GPIO4_IO13 | PC,
		.gp = IMX_GPIO_NR(4, 13),
	},
};
#endif

int dram_init(void)
{
	gd->ram_size = imx_ddr_size();

	return 0;
}

static iomux_v3_cfg_t const wdog_pads[] = {
	MX7D_PAD_GPIO1_IO00__WDOG1_WDOG_B | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const uart5_pads[] = {
	MX7D_PAD_GPIO1_IO06__UART5_RX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX7D_PAD_GPIO1_IO07__UART5_TX_DATA | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX7D_PAD_GPIO1_IO04__UART5_CTS_B | MUX_PAD_CTRL(UART_PAD_CTRL),
	MX7D_PAD_GPIO1_IO05__UART5_RTS_B | MUX_PAD_CTRL(UART_PAD_CTRL),
};

static iomux_v3_cfg_t const usdhc1_pads[] = {
	MX7D_PAD_SD1_CLK__SD1_CLK | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD1_CMD__SD1_CMD | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD1_DATA0__SD1_DATA0 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD1_DATA1__SD1_DATA1 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD1_DATA2__SD1_DATA2 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD1_DATA3__SD1_DATA3 | MUX_PAD_CTRL(USDHC_PAD_CTRL),

	MX7D_PAD_SD1_CD_B__GPIO5_IO0 | MUX_PAD_CTRL(USDHC_PAD_CTRL_CD),
	MX7D_PAD_SD1_WP__GPIO5_IO1 | MUX_PAD_CTRL(USDHC_PAD_CTRL_WP),
	MX7D_PAD_SD1_RESET_B__GPIO5_IO2 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const usdhc3_emmc_pads[] = {
	MX7D_PAD_SD3_CLK__SD3_CLK | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_CMD__SD3_CMD | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA0__SD3_DATA0 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA1__SD3_DATA1 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA2__SD3_DATA2 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA3__SD3_DATA3 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA4__SD3_DATA4 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA5__SD3_DATA5 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA6__SD3_DATA6 | MUX_PAD_CTRL(USDHC_PAD_CTRL),
	MX7D_PAD_SD3_DATA7__SD3_DATA7 | MUX_PAD_CTRL(USDHC_PAD_CTRL),

	MX7D_PAD_SD3_RESET_B__GPIO6_IO11 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const gpio_aiot_g3_led_pads[] = {
	MX7D_PAD_LCD_DATA06__GPIO3_IO11 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_LCD_DATA05__GPIO3_IO10 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_LCD_DATA07__GPIO3_IO12 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_LCD_DATA08__GPIO3_IO13 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const gpio_aiot_g3l_led_pads[] = {
	MX7D_PAD_LCD_DATA05__GPIO3_IO10 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_LCD_DATA06__GPIO3_IO11 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_LCD_DATA07__GPIO3_IO12 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const gpio_ax1_led_pads[] = {
	MX7D_PAD_EPDC_DATA07__GPIO2_IO7 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

#define MAINTENANCE_MODE_GPIO IMX_GPIO_NR(1, 9)

static iomux_v3_cfg_t const maintenance_mode_gpio_pads[] = {
	MX7D_PAD_GPIO1_IO09__GPIO1_IO9 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

#define G_OTG1_PWREN IMX_GPIO_NR(4, 15)
static iomux_v3_cfg_t const usb_otg1_pads[] = {
	MX7D_PAD_I2C4_SDA__GPIO4_IO15 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

static iomux_v3_cfg_t const x1_addon_pwr_en[] = {
	MX7D_PAD_EPDC_DATA04__GPIO2_IO4 | MUX_PAD_CTRL(NO_PAD_CTRL),
};


#define IOX_SDI IMX_GPIO_NR(1, 9)
#define IOX_STCP IMX_GPIO_NR(1, 12)
#define IOX_SHCP IMX_GPIO_NR(1, 13)

static iomux_v3_cfg_t const iox_pads[] = {
	/* IOX_SDI */
	MX7D_PAD_GPIO1_IO09__GPIO1_IO9	| MUX_PAD_CTRL(NO_PAD_CTRL),
	/* IOX_STCP */
	MX7D_PAD_GPIO1_IO12__GPIO1_IO12	| MUX_PAD_CTRL(NO_PAD_CTRL),
	/* IOX_SHCP */
	MX7D_PAD_GPIO1_IO13__GPIO1_IO13	| MUX_PAD_CTRL(NO_PAD_CTRL),
};

#define AX1_PCIE_CLKREQ IMX_GPIO_NR(2, 12)
#define AX1_PCIE_WAKE IMX_GPIO_NR(6, 14)

static iomux_v3_cfg_t const ax1_pcie_pads[] = {
	/* G_PCIE_CLKREQ */
	MX7D_PAD_EPDC_DATA12__GPIO2_IO12	| MUX_PAD_CTRL(NO_PAD_CTRL),
	/* G_PCIE_WAKE */
	MX7D_PAD_SAI1_TX_SYNC__GPIO6_IO14	| MUX_PAD_CTRL(NO_PAD_CTRL),
};

/*
 * PCIE_DIS_B --> Q0
 * PCIE_RST_B --> Q1
 * HDMI_RST_B --> Q2
 * PERI_RST_B --> Q3
 * SENSOR_RST_B --> Q4
 * ENET_RST_B --> Q5
 * PERI_3V3_EN --> Q6
 * LCD_PWR_EN --> Q7
 */
enum qn {
	PCIE_DIS_B,
	PCIE_RST_B,
	HDMI_RST_B,
	PERI_RST_B,
	SENSOR_RST_B,
	ENET_RST_B,
	PERI_3V3_EN,
	LCD_PWR_EN,
};

enum qn_func {
	qn_reset,
	qn_enable,
	qn_disable,
};

enum qn_level {
	qn_low = 0,
	qn_high = 1,
};

static enum qn_level seq[3][2] = {
	{0, 1}, {1, 1}, {0, 0}
};

static enum qn_func qn_output[8] = {
	qn_disable, qn_reset, qn_reset, qn_reset, qn_reset, qn_reset, qn_enable,
	qn_disable
};

static void iox74lv_init(void)
{
	int i;

	for (i = 7; i >= 0; i--) {
		gpio_direction_output(IOX_SHCP, 0);
		gpio_direction_output(IOX_SDI, seq[qn_output[i]][0]);
		udelay(500);
		gpio_direction_output(IOX_SHCP, 1);
		udelay(500);
	}

	gpio_direction_output(IOX_STCP, 0);
	udelay(500);
	/*
	  * shift register will be output to pins
	  */
	gpio_direction_output(IOX_STCP, 1);

	for (i = 7; i >= 0; i--) {
		gpio_direction_output(IOX_SHCP, 0);
		gpio_direction_output(IOX_SDI, seq[qn_output[i]][1]);
		udelay(500);
		gpio_direction_output(IOX_SHCP, 1);
		udelay(500);
	}
	gpio_direction_output(IOX_STCP, 0);
	udelay(500);
	/*
	  * shift register will be output to pins
	  */
	gpio_direction_output(IOX_STCP, 1);
};

#ifdef CONFIG_NAND_MXS
static iomux_v3_cfg_t const gpmi_pads[] = {
	MX7D_PAD_SD3_DATA0__NAND_DATA00 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA1__NAND_DATA01 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA2__NAND_DATA02 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA3__NAND_DATA03 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA4__NAND_DATA04 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA5__NAND_DATA05 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA6__NAND_DATA06 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_DATA7__NAND_DATA07 | MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_CLK__NAND_CLE	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_CMD__NAND_ALE	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_STROBE__NAND_RE_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SD3_RESET_B__NAND_WE_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_MCLK__NAND_WP_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_RX_BCLK__NAND_CE3_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_RX_SYNC__NAND_CE2_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_RX_DATA__NAND_CE1_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_TX_BCLK__NAND_CE0_B	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_TX_SYNC__NAND_DQS	| MUX_PAD_CTRL(NAND_PAD_CTRL),
	MX7D_PAD_SAI1_TX_DATA__NAND_READY_B	| MUX_PAD_CTRL(NAND_PAD_READY0_CTRL),
};

static void setup_gpmi_nand(void)
{
	imx_iomux_v3_setup_multiple_pads(gpmi_pads, ARRAY_SIZE(gpmi_pads));

	/* NAND_USDHC_BUS_CLK is set in rom */
	set_clk_nand();
}
#endif

#ifdef CONFIG_FEC_MXC
static iomux_v3_cfg_t const fec2_pads[] = {
	MX7D_PAD_EPDC_SDCE0__ENET2_RGMII_RX_CTL | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDCLK__ENET2_RGMII_RD0 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDLE__ENET2_RGMII_RD1 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDOE__ENET2_RGMII_RD2 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDSHR__ENET2_RGMII_RD3 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE1__ENET2_RGMII_RXC | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),

	MX7D_PAD_EPDC_GDRL__ENET2_RGMII_TX_CTL | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE2__ENET2_RGMII_TD0 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE3__ENET2_RGMII_TD1 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_GDCLK__ENET2_RGMII_TD2 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_GDOE__ENET2_RGMII_TD3 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_GDSP__ENET2_RGMII_TXC | MUX_PAD_CTRL(ENET_PAD_CTRL),

	MX7D_PAD_GPIO1_IO14__ENET2_MDIO | MUX_PAD_CTRL(ENET_PAD_CTRL_MII),
	MX7D_PAD_GPIO1_IO15__ENET2_MDC | MUX_PAD_CTRL(ENET_PAD_CTRL_MII),

	/* GPIO_LAN1_RST_N */
	MX7D_PAD_EPDC_BDR1__GPIO2_IO29 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_EPDC_PWR_COM__GPIO2_IO30,	/* GPIO_LAN1_COMA_MODE */
};

static iomux_v3_cfg_t const fec2_pads_g3l[] = {
	MX7D_PAD_GPIO1_IO15__ENET2_MDC | MUX_PAD_CTRL(ENET_PAD_CTRL_MII_G3L),
	MX7D_PAD_GPIO1_IO14__ENET2_MDIO | MUX_PAD_CTRL(ENET_PAD_CTRL_MII_G3L),

	MX7D_PAD_EPDC_GDRL__ENET2_RGMII_TX_CTL | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE3__ENET2_RGMII_TD1 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE2__ENET2_RGMII_TD0 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_EPDC_SDLE__ENET2_RGMII_RD1 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDCLK__ENET2_RGMII_RD0 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_SDCE1__ENET2_RX_ER | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_PWR_COM__ENET2_CRS | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_EPDC_BDR0__CCM_ENET_REF_CLK2 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),

	/* GPIO_LAN1_RST_N */
	MX7D_PAD_EPDC_BDR1__GPIO2_IO29 | MUX_PAD_CTRL(NO_PAD_CTRL),
};

#ifdef CONFIG_X1_2ND_ETHERNET
static iomux_v3_cfg_t const fec1_pads[] = {
	MX7D_PAD_ENET1_RGMII_RX_CTL__ENET1_RGMII_RX_CTL | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_RD0__ENET1_RGMII_RD0 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_RD1__ENET1_RGMII_RD1 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_RD2__ENET1_RGMII_RD2 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_RD3__ENET1_RGMII_RD3 | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_RXC__ENET1_RGMII_RXC | MUX_PAD_CTRL(ENET_RX_PAD_CTRL),

	MX7D_PAD_ENET1_RGMII_TX_CTL__ENET1_RGMII_TX_CTL | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_TD0__ENET1_RGMII_TD0 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_TD1__ENET1_RGMII_TD1 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_TD2__ENET1_RGMII_TD2 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_TD3__ENET1_RGMII_TD3 | MUX_PAD_CTRL(ENET_PAD_CTRL),
	MX7D_PAD_ENET1_RGMII_TXC__ENET1_RGMII_TXC | MUX_PAD_CTRL(ENET_PAD_CTRL),

	MX7D_PAD_SD2_CD_B__ENET1_MDIO | MUX_PAD_CTRL(ENET_PAD_CTRL_MII),
	MX7D_PAD_SD2_WP__ENET1_MDC | MUX_PAD_CTRL(ENET_PAD_CTRL_MII),

	/* G_ENET1_NRESET */
	MX7D_PAD_ENET1_RX_CLK__GPIO7_IO13 | MUX_PAD_CTRL(NO_PAD_CTRL),
	MX7D_PAD_ENET1_CRS__GPIO7_IO14,	/* G_ENET1_COMA_MD */
};
#endif

static void setup_iomux_fec(void)
{
	switch (x1_board_info.board_type) {
	case BOARDTYPE_AX1:
	case BOARDTYPE_AIOTG_G3:
	case BOARDTYPE_AIOTG_G3_D2:
	case BOARDTYPE_AIOTG_G3_W1:
	case BOARDTYPE_AIOTG_G3_M1:
	case BOARDTYPE_AIOTG_G3_W2:
		imx_iomux_v3_setup_multiple_pads(fec2_pads, ARRAY_SIZE(fec2_pads));
		break;
	case BOARDTYPE_AIOTG_G3L:
		imx_iomux_v3_setup_multiple_pads(fec2_pads_g3l, ARRAY_SIZE(fec2_pads_g3l));
		break;
	default:
		break;
	}

#ifdef CONFIG_X1_2ND_ETHERNET
	imx_iomux_v3_setup_multiple_pads(fec1_pads, ARRAY_SIZE(fec1_pads));
#endif
}
#endif

static void setup_iomux_uart(void)
{
	imx_iomux_v3_setup_multiple_pads(uart5_pads, ARRAY_SIZE(uart5_pads));
}

#ifdef CONFIG_FSL_ESDHC

#define USDHC1_CD_GPIO	IMX_GPIO_NR(5, 0)
#define USDHC1_PWR_GPIO	IMX_GPIO_NR(5, 2)
#define USDHC3_PWR_GPIO IMX_GPIO_NR(6, 11)

static struct fsl_esdhc_cfg usdhc_cfg[3] = {
	{USDHC1_BASE_ADDR, 0, 4},
	{USDHC3_BASE_ADDR},
};

int board_mmc_get_env_dev(int devno)
{
	if (devno == 2)
		devno--;

	return devno;
}

int board_mmc_getcd(struct mmc *mmc)
{
	struct fsl_esdhc_cfg *cfg = (struct fsl_esdhc_cfg *)mmc->priv;
	int ret = 0;

	switch (cfg->esdhc_base) {
	case USDHC1_BASE_ADDR:
#ifndef CONFIG_SPL_BUILD
		if (x1_board_info.exist_sdhc1)
#endif
			ret = !gpio_get_value(USDHC1_CD_GPIO);
		break;
	case USDHC3_BASE_ADDR:
		ret = 1; /* Assume uSDHC3 emmc is always present */
		break;
	}

	return ret;
}

int board_mmc_init(bd_t *bis)
{
	int i, ret;
	/*
	 * According to the board_mmc_init() the following map is done:
	 * (U-Boot device node)    (Physical Port)
	 * mmc0                    USDHC1
	 * mmc2                    USDHC3 (eMMC)
	 */
	for (i = 0; i < CONFIG_SYS_FSL_USDHC_NUM; i++) {
		switch (i) {
		case 0:
#ifndef CONFIG_SPL_BUILD
			if (x1_board_info.exist_sdhc1) {
#endif
				imx_iomux_v3_setup_multiple_pads(
					usdhc1_pads, ARRAY_SIZE(usdhc1_pads));
				gpio_request(USDHC1_CD_GPIO, "usdhc1_cd");
				gpio_direction_input(USDHC1_CD_GPIO);
				gpio_request(USDHC1_PWR_GPIO, "usdhc1_pwr");
				gpio_direction_output(USDHC1_PWR_GPIO, 0);
				udelay(500);
				gpio_direction_output(USDHC1_PWR_GPIO, 1);
				usdhc_cfg[0].sdhc_clk = mxc_get_clock(MXC_ESDHC_CLK);
				break;
#ifndef CONFIG_SPL_BUILD
			}
#endif
			continue;
		case 1:
			imx_iomux_v3_setup_multiple_pads(
				usdhc3_emmc_pads, ARRAY_SIZE(usdhc3_emmc_pads));
			gpio_request(USDHC3_PWR_GPIO, "usdhc3_pwr");
			gpio_direction_output(USDHC3_PWR_GPIO, 0);
			udelay(500);
			gpio_direction_output(USDHC3_PWR_GPIO, 1);
			usdhc_cfg[1].sdhc_clk = mxc_get_clock(MXC_ESDHC3_CLK);
			break;
		default:
			printf("Warning: you configured more USDHC controllers"
				"(%d) than supported by the board\n", i + 1);
			return -EINVAL;
		}

		ret = fsl_esdhc_initialize(bis, &usdhc_cfg[i]);
		if (ret)
			return ret;
	}

	return 0;
}

#endif

#ifdef CONFIG_FEC_MXC
#define PHY_RST_N_G3L	IMX_GPIO_NR(2, 29)
int board_eth_init(bd_t *bis)
{
	int ret;

	if ((x1_board_info.board_type == BOARDTYPE_UNKNOWN) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z))
		return 0;

	setup_iomux_fec();

	ret = fecmxc_initialize_multi(bis, 0,
		CONFIG_FEC_MXC_PHYADDR, IMX_FEC_BASE);
	if (ret)
		printf("FEC2 MXC: %s:failed\n", __func__);

#ifdef CONFIG_X1_2ND_ETHERNET
	ret = fecmxc_initialize_multi(bis, 1,
		CONFIG_FEC_MXC_PHYADDR, ENET_IPS_BASE_ADDR);
	if (ret) {
		printf("FEC1 MXC: %s:failed\n", __func__);
		return ret;
	}
#endif

	return ret;
}

static int enable_pll_enet100base(void)
{
	u32 reg;
	s32 timeout = 100000;

	struct mxc_ccm_anatop_reg *ccm_anatop = (struct mxc_ccm_anatop_reg *)
						 ANATOP_BASE_ADDR;

	reg = readl(&ccm_anatop->pll_enet);
	/* If pll_enet powered up, no need to set it again */
	if (reg & ANADIG_PLL_ENET_PWDN_MASK) {
		reg &= ~ANADIG_PLL_ENET_PWDN_MASK;
		writel(reg, &ccm_anatop->pll_enet);

		while (timeout--) {
			if (readl(&ccm_anatop->pll_enet) & ANADIG_PLL_LOCK)
				break;
		}

		if (timeout <= 0) {
			/* If timeout, we set pwdn for pll_enet. */
			reg |= ANADIG_PLL_ENET_PWDN_MASK;
			return -ETIME;
		}
	}

	/* Clear bypass */
	writel(CCM_ANALOG_PLL_ENET_BYPASS_MASK, &ccm_anatop->pll_enet_clr);

	writel((CCM_ANALOG_PLL_ENET_ENABLE_CLK_500MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_250MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_125MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_100MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_50MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_40MHZ_MASK
		| CCM_ANALOG_PLL_ENET_ENABLE_CLK_25MHZ_MASK),
	       &ccm_anatop->pll_enet_set);

	return 0;
}

static int set_clk_enet100base(void)
{
	u32 target;
	int ret;
	u32 enet1_ref, enet2_ref;

	/* disable the clock first */
	clock_enable(CCGR_ENET1, 0);
	clock_enable(CCGR_ENET2, 0);

	enet1_ref = ENET1_REF_CLK_ROOT_FROM_PLL_ENET_MAIN_50M_CLK;
	enet2_ref = ENET1_REF_CLK_ROOT_FROM_PLL_ENET_MAIN_50M_CLK;

	ret = enable_pll_enet100base();
	if (ret != 0)
		return ret;

	/* set enet axi clock 49M: 392/8 */
	target = CLK_ROOT_ON | ENET_AXI_CLK_ROOT_FROM_PLL_SYS_PFD4_CLK |
		 CLK_ROOT_PRE_DIV(CLK_ROOT_PRE_DIV2) |
		 CLK_ROOT_POST_DIV(CLK_ROOT_POST_DIV4);
	clock_set_target_val(ENET_AXI_CLK_ROOT, target);

	target = CLK_ROOT_ON | enet1_ref |
		 CLK_ROOT_PRE_DIV(CLK_ROOT_PRE_DIV1) |
		 CLK_ROOT_POST_DIV(CLK_ROOT_POST_DIV1);
	clock_set_target_val(ENET1_REF_CLK_ROOT, target);

	target = CLK_ROOT_ON | ENET1_TIME_CLK_ROOT_FROM_PLL_ENET_MAIN_100M_CLK |
		 CLK_ROOT_PRE_DIV(CLK_ROOT_PRE_DIV1) |
		 CLK_ROOT_POST_DIV(CLK_ROOT_POST_DIV4);
	clock_set_target_val(ENET1_TIME_CLK_ROOT, target);

	target = CLK_ROOT_ON | enet2_ref |
		 CLK_ROOT_PRE_DIV(CLK_ROOT_PRE_DIV1) |
		 CLK_ROOT_POST_DIV(CLK_ROOT_POST_DIV1);
	clock_set_target_val(ENET2_REF_CLK_ROOT, target);

	target = CLK_ROOT_ON | ENET2_TIME_CLK_ROOT_FROM_PLL_ENET_MAIN_100M_CLK |
		 CLK_ROOT_PRE_DIV(CLK_ROOT_PRE_DIV1) |
		 CLK_ROOT_POST_DIV(CLK_ROOT_POST_DIV4);
	clock_set_target_val(ENET2_TIME_CLK_ROOT, target);

	/* enable clock */
	clock_enable(CCGR_ENET1, 1);
	clock_enable(CCGR_ENET2, 1);

	return 0;
}

static int setup_fec(void)
{
	struct iomuxc_gpr_base_regs *const iomuxc_gpr_regs
		= (struct iomuxc_gpr_base_regs *) IOMUXC_GPR_BASE_ADDR;

	/* Use 125M anatop REF_CLK1 for ENET1, clear gpr1[13], gpr1[17]*/
	clrsetbits_le32(&iomuxc_gpr_regs->gpr[1],
		(IOMUXC_GPR_GPR1_GPR_ENET1_TX_CLK_SEL_MASK |
		 IOMUXC_GPR_GPR1_GPR_ENET1_CLK_DIR_MASK), 0);

	if (x1_board_info.board_type == BOARDTYPE_AIOTG_G3L)
		return set_clk_enet100base();

	return set_clk_enet(ENET_125MHz);
}

#define VSC_PHY_MAIN_PAGE	(0x00)
#define VSC_PHY_EXT1_PAGE	(0x01)
#define VSC_PHY_EXT2_PAGE	(0x02)
#define VSC_PHY_EXT3_PAGE	(0x03)
#define VSC_PHY_EXT4_PAGE	(0x04)
#define VSC_PHY_GPIO_PAGE	(0x10)
#define VSC_PHY_TEST_PAGE	(0x2A30)
#define VSC_PHY_TR_PAGE		(0x52B5)

static int vsc_phy_read(struct phy_device *phydev, int page,
			int regnum)
{
	int ret;

	/* set access page */
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1f, page);

	/* read val */
	ret = phy_read(phydev, MDIO_DEVAD_NONE, regnum);

	if (page != VSC_PHY_MAIN_PAGE) {
		phy_write(phydev, MDIO_DEVAD_NONE, 0x1f,
			  VSC_PHY_MAIN_PAGE);
	}

	return ret;
}

static int vsc_phy_write(struct phy_device *phydev, int page,
			 int regnum, u16 val)
{
	int ret;

	/* set access page */
	phy_write(phydev, MDIO_DEVAD_NONE, 0x1f, page);

	/* write val */
	ret = phy_write(phydev, MDIO_DEVAD_NONE, regnum, val);

	if (page != VSC_PHY_MAIN_PAGE) {
		phy_write(phydev, MDIO_DEVAD_NONE, 0x1f,
			  VSC_PHY_MAIN_PAGE);
	}

	return ret;
}

static int vsc_phy_write_masked(struct phy_device *phydev, int page,
				int regnum, u16 val, u16 mask)
{
	int tmp;

	tmp = vsc_phy_read(phydev, page, regnum);
	tmp &= ~mask;
	tmp |= (val & mask);

	return vsc_phy_write(phydev, page, regnum, tmp);
}

static const u16 init_script_nano_rev_a[] = {
//     Op,   Page,  Reg, Value,  Mask
//      0,      1,    2,     3,     4
//     --, ------, ----, ------, -----
        0, 0x0000, 0x1f, 0x0000, 0xffff,
        1, 0x0000, 0x16, 0x0001, 0x0001,
        0, 0x0001, 0x1f, 0x2A30, 0xffff,
        1, 0x2A30, 0x18, 0x0000, 0x0400,
        1, 0x2A30, 0x05, 0x0c00, 0x0e00,
        1, 0x2A30, 0x08, 0x8000, 0x8000,
        0, 0x2A30, 0x1f, 0x52B5, 0xffff,
        0, 0x52B5, 0x12, 0x0068, 0xffff,
        0, 0x52B5, 0x11, 0x8980, 0xffff,
        0, 0x52B5, 0x10, 0x8f90, 0xffff,
        0, 0x52B5, 0x12, 0x0000, 0xffff,
        0, 0x52B5, 0x11, 0x0003, 0xffff,
        0, 0x52B5, 0x10, 0x8796, 0xffff,
        0, 0x52B5, 0x12, 0x0050, 0xffff,
        0, 0x52B5, 0x11, 0x100f, 0xffff,
        0, 0x52B5, 0x10, 0x87fa, 0xffff,
        0, 0x52B5, 0x1f, 0x2A30, 0xffff,
        1, 0x2A30, 0x08, 0x0000, 0x8000,
        0, 0x2A30, 0x1f, 0x0000, 0xffff,
        1, 0x0000, 0x16, 0x0000, 0x0001,
	0xf, 0xffff, 0xff, 0xffff, 0xffff,
};

static int load_init_script(struct phy_device *phydev)
{
	u16 op, page, reg, value, mask;
	int i;

	for (i = 0; i < ARRAY_SIZE(init_script_nano_rev_a);) {
		op    = init_script_nano_rev_a[i++];
		page  = init_script_nano_rev_a[i++];
		reg   = init_script_nano_rev_a[i++];
		value = init_script_nano_rev_a[i++];
		mask  = init_script_nano_rev_a[i++];

		if ((op == 0) || (op == 1)) {
			vsc_phy_write_masked(phydev, page, reg,
					     value, mask);
		}
	}

	return 0;
}

static int apply_nano_patch(struct phy_device *phydev)
{
	int val;

	val = vsc_phy_read(phydev, VSC_PHY_EXT1_PAGE, 0x17);
	if ((val & 0xf800) >> 11 != 0) {
		printf("unsupported port %d\n", val);
		return -1;
	}

	return load_init_script(phydev);
}


struct vsc8501_gpios {
	unsigned nreset;
	unsigned coma_mode;
};

static struct vsc8501_gpios fec2_vsc8501_gpios = {
	.nreset         = IMX_GPIO_NR(2, 29),
	.coma_mode      = IMX_GPIO_NR(2, 30),
};

#ifdef CONFIG_X1_2ND_ETHERNET
static struct vsc8501_gpios fec1_vsc8501_gpios = {
	.nreset		= IMX_GPIO_NR(7, 13),
	.coma_mode	= IMX_GPIO_NR(7, 14),
};
#endif

static int vsc8501_phy_config(struct phy_device *phydev, struct vsc8501_gpios *gpios)
{
	int ret;

	/* enable rgmii rxc/gtxc skew and phy mode select to RGMII copper */

	gpio_request(gpios->coma_mode, "lan1_coma_mode");

	/* Reset active, drive low. */
	gpio_direction_output(gpios->nreset, 0);
	mdelay(2);

	/* 1. COMA_MODE active, drive high. */
	gpio_direction_output(gpios->coma_mode, 1);
	mdelay(10);

	/* 2. Apply power. */
	/* 3. Apply RefClk. */

	/* 4. Release reset, drive high. */
	gpio_direction_output(gpios->nreset, 1);
	mdelay(105);

	/* 5. Wait 120 ms minimum. */
	mdelay(120);

	/* 6. Apply patch from PHY_API. */
	ret = apply_nano_patch(phydev);
	if (ret)
		return ret;

	/* 7. Configure register 23 for MAC interface mode. */
	/* 8. Write new register 23. */
	vsc_phy_write_masked(phydev, VSC_PHY_MAIN_PAGE, 0x17,
			     0x3000, 0x3000);	/* RGMII */

	/* 9. Software reset. */
	/* 10. Write new register 0. */
	/* 11. Read register 0 untio bit 15 equals to 0. */
	ret = phy_reset(phydev);
	if (ret)
		return ret;

	/* 12. For RGMII mode: Configure register 20E2. */
	vsc_phy_write_masked(phydev, VSC_PHY_EXT2_PAGE, 0x14,
			     0x0000, 0x0800);
	vsc_phy_write_masked(phydev, VSC_PHY_EXT2_PAGE, 0x14,
			     0x0033, 0x0077);

	/* 13. Release the COMA_MODE pin. */
	gpio_direction_output(gpios->coma_mode, 0);

	/* LED configuration */
	vsc_phy_write_masked(phydev, VSC_PHY_MAIN_PAGE, 0x1d,
			     0x0ee0, 0x0fff);
	vsc_phy_write_masked(phydev, VSC_PHY_GPIO_PAGE, 0x0e,
			     0x0000, 0x0200);
	vsc_phy_write_masked(phydev, VSC_PHY_GPIO_PAGE, 0x1e,
			     0x0800, 0x0c00);

	return 0;
}

int board_phy_config(struct phy_device *phydev)
{
	struct vsc8501_gpios *gpios;
	int ret;

	gpios = &fec2_vsc8501_gpios;
#ifdef CONFIG_X1_2ND_ETHERNET
	if (phydev->dev->index == 1)
		gpios = &fec1_vsc8501_gpios;
#endif

	if ((x1_board_info.board_type == BOARDTYPE_UNKNOWN) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z))
		return 0;

	if ((x1_board_info.board_type == BOARDTYPE_AX1) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_D2) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_W1) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_M1) ||
	    (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_W2)) {
		ret = vsc8501_phy_config(phydev, gpios);
		if (ret)
			return ret;
	}

	if (phydev->drv->config)
		phydev->drv->config(phydev);
	return 0;
}
#endif

static void show_boot_source(void)
{
	enum boot_device boot_dev;

	puts("Boot Source: ");

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case SD1_BOOT:
		puts("SD\n");
		break;
	case MMC3_BOOT:
		puts("eMMC\n");
		break;
	case QSPI_BOOT:
		puts("QSPI Flash\n");
		break;
	default:
		puts("unknown\n");
		break;
	}
}

#ifndef CONFIG_SPL_BUILD
static enum env_location env_locations[] = {
	ENVL_MMC,
	ENVL_SPI_FLASH,
};

enum env_location env_get_location(enum env_operation op, int prio)
{
	enum boot_device boot_dev;

	if (prio >= ARRAY_SIZE(env_locations))
		return ENVL_UNKNOWN;

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case MMC3_BOOT:
	case SD1_BOOT:
		return ENVL_MMC;
	case QSPI_BOOT:
		return ENVL_SPI_FLASH;
	default:
		return ENVL_UNKNOWN;
	}
}
#endif

#ifdef CONFIG_SYS_MMC_ENV_PART
uint mmc_get_env_part(struct mmc *mmc)
{
	enum boot_device boot_dev;

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case MMC3_BOOT:
		/* Support only first boot partition. */
		return 1;
	case SD1_BOOT:
	default:
		return CONFIG_SYS_MMC_ENV_PART;
	}
}
#endif

#define BMIC_I2C_BUS		1
#define BMIC_I2C_GENERAL_SLAVE	0x10
#define BMIC_I2C_GENERAL_VER	0x01
#define BMIC_I2C_EEPROM_SLAVE	0x15
#define BMIC_I2C_EEPROM_VER	0x01
#define BMIC_I2C_EEPROM_DATA	0x02
#define BMIC_I2C_WAKE_MAIN	0xab
#define BMIC_I2C_EEPROM_SIZE	1024

/*
 * U-boot must not access to BMIC for about 20 ms
 * after it woke up.
 * Because BMIC enters to LPTMR interrupt handler for 7 ms
 * after it woke up after 10 ms.
 */
#define BMIC_I2C_WAIT_TIME_AFTER_STOP_MODE_MS	(20)

static void x1l_bmic_wake_main(void)
{
	u8 wake_main_req[] = { 0x51, 0x00, 0x00, 0x00 };
	i2c_write(BMIC_I2C_GENERAL_SLAVE, BMIC_I2C_WAKE_MAIN, 1,
		  wake_main_req, sizeof(wake_main_req));
}

static int bmic_is_running(void)
{
	/*
	 * Read BMIC general version to check it is running
	 * and wake it up by I2C address match if it is stop mode.
	 */
	u8 dummy[2];
	return !!i2c_read(BMIC_I2C_GENERAL_SLAVE, BMIC_I2C_GENERAL_VER, 1, dummy, sizeof(dummy));
}

static void i2c_force_idle_bus(struct i2c_pads_info *pins);
static void wait_for_bmic_accessible(void)
{
	int count = 0;

	puts("BMIC is not running. Waiting.");
	while (bmic_is_running()) {
		/* Print dot as indicator at 500 ms interval.*/
		if (!(count++ % (500 / BMIC_I2C_WAIT_TIME_AFTER_STOP_MODE_MS)))
			putc('.');

		mdelay(BMIC_I2C_WAIT_TIME_AFTER_STOP_MODE_MS);
		i2c_force_idle_bus(&i2c_pad_info4);
		setup_i2c(0, CONFIG_SYS_I2C_SPEED, 0x7f, &i2c_pad_info4);
	}
	puts("done.\n");
}

/* wakeup when bmic is sleep mode. */
static void bmic_init(void)
{
	i2c_set_bus_num(BMIC_I2C_BUS);

	i2c_init(CONFIG_SYS_I2C_SPEED, BMIC_I2C_GENERAL_SLAVE);

	x1l_bmic_wake_main();
	mdelay(20);

	i2c_probe(BMIC_I2C_GENERAL_SLAVE);
	mdelay(20);

	i2c_probe(BMIC_I2C_GENERAL_SLAVE);
	mdelay(20);
}

static int bmic_eeprom_init(void)
{
	i2c_set_bus_num(BMIC_I2C_BUS);

	i2c_init(CONFIG_SYS_I2C_SPEED, BMIC_I2C_EEPROM_SLAVE);

	return i2c_probe(BMIC_I2C_EEPROM_SLAVE);
}

static int bmic_eeprom_read_byte_internal(u16 addr, u8 *data)
{
	if (i2c_write(BMIC_I2C_EEPROM_SLAVE, BMIC_I2C_EEPROM_DATA, 1, (u8 *)&addr, sizeof(addr))) {
		printf("bmic eeprom read error! addr = %d: Command Write\n", addr);
		return -1;
	}

	if (i2c_read(BMIC_I2C_EEPROM_SLAVE, 0, 0, data, 1)) {
		printf("bmic eeprom read error! addr = %d: Data Read\n", addr);
		return -1;
	}

	return 0;
}

static int bmic_eeprom_read_byte(u16 addr, u8 *data)
{
	while (bmic_eeprom_read_byte_internal(addr, data))
		wait_for_bmic_accessible();

	return 0;
}

static int bmic_eeprom_read(u16 addr, void *data, u16 size)
{
	u8 *pdata = (u8 *)data;

	if ((addr + size) > BMIC_I2C_EEPROM_SIZE) {
		printf("bmic eeprom read error! Size \n");
		return -1;
	}

	while (size--) {
		if (bmic_eeprom_read_byte(addr, pdata) != 0)
			return -1;
		addr++;
		pdata++;
	}

	return 0;
}

#define SVALID_INVALID		0x00
#define SVALID_SERIALNO		0x01
#define SVALID_KOTAI		0x02

int board_info_init(void)
{
	memcpy(&x1_board_info, 0x00, sizeof(x1_board_info));

	bmic_init();

	if (bmic_eeprom_init()) {
		printf("error: BMIC EEPROM detect\n");
		x1_board_info.board_type = BOARDTYPE_UNKNOWN;
		x1_board_info.revision = 0x0000;
		x1_board_info.serial_no = 0;
		return -1;
	}

	bmic_eeprom_read(0, &x1_board_info.board_type, sizeof(x1_board_info.board_type));
	bmic_eeprom_read(4, &x1_board_info.revision, sizeof(x1_board_info.revision));
	bmic_eeprom_read(15, &x1_board_info.serial_no_valid, sizeof(x1_board_info.serial_no_valid));
	if (x1_board_info.serial_no_valid == SVALID_SERIALNO)
		bmic_eeprom_read(8, &x1_board_info.serial_no, sizeof(x1_board_info.serial_no));
	else if(x1_board_info.serial_no_valid == SVALID_KOTAI)
		bmic_eeprom_read(8, &x1_board_info.kotai, sizeof(x1_board_info.kotai));

	bmic_eeprom_read(16, &x1_board_info.dram, sizeof(x1_board_info.dram));
	bmic_eeprom_read(20, &x1_board_info.xtal, sizeof(x1_board_info.xtal));
	bmic_eeprom_read(32, &x1_board_info.usermac, 6);

	puts("Board Type: ");
	switch (x1_board_info.board_type) {
	case BOARDTYPE_AX1:
		puts("Armadillo-X1");
		x1_board_info.exist_sdhc1 = 0;
		break;
	case BOARDTYPE_AIOTG_G3L:
		puts("Armadillo-IoT G3L");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3:
		puts("Armadillo-IoT G3");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3_D2:
		puts("Armadillo-IoT G3 D2");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3_W1:
		puts("Armadillo-IoT G3 W1");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3_M1:
		puts("Armadillo-IoT G3 M1");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z:
		puts("Armadillo-IoT G3 D2(AGX3031-E00Z)");
		x1_board_info.exist_sdhc1 = 1;
		break;
	case BOARDTYPE_AIOTG_G3_W2:
		puts("Armadillo-IoT G3 W2");
		x1_board_info.exist_sdhc1 = 1;
		break;
	default:
		printf("Unknown Board Type\n");
		x1_board_info.board_type = BOARDTYPE_UNKNOWN;
		x1_board_info.exist_sdhc1 = 1;
		break;
	}
	printf("(%08x)\n", x1_board_info.board_type);

	printf("Revision: %04x\n", x1_board_info.revision);

	if (x1_board_info.serial_no_valid == SVALID_SERIALNO)
		printf("S/N: %d\n", (int)x1_board_info.serial_no);
	else if(x1_board_info.serial_no_valid == SVALID_KOTAI)
		printf("S/N: %04X%04d%04d\n", (int)x1_board_info.kotai.product_id,
					(int)x1_board_info.kotai.lot, (int)x1_board_info.kotai.number);

	printf("DRAM: %08x\n", x1_board_info.dram);
	printf("XTAL: %02x\n", x1_board_info.xtal);

	return 0;
}

static void i2c_force_idle_bus(struct i2c_pads_info *pins)
{
	int i;

	gpio_direction_input(pins->sda.gp);
	gpio_direction_input(pins->scl.gp);

	imx_iomux_v3_setup_pad(pins->sda.gpio_mode);
	imx_iomux_v3_setup_pad(pins->scl.gpio_mode);

	/* send start/stop bit */
	udelay(50);
	gpio_direction_output(pins->sda.gp, 0);
	udelay(50);
	gpio_direction_input(pins->sda.gp);
	udelay(50);

	/* Send high and low on the SCL line */
	for (i = 0; i < 54; i++) {
		gpio_direction_output(pins->scl.gp, 0);
		udelay(50);
		gpio_direction_input(pins->scl.gp);
		udelay(50);
	}

	/* send stop bit */
	gpio_direction_output(pins->scl.gp, 0);
	gpio_direction_output(pins->sda.gp, 0);
	udelay(50);
	gpio_direction_input(pins->scl.gp);
	udelay(50);
	gpio_direction_input(pins->sda.gp);
}

#define X1_ADDON_EEPROM_SLAVE	0x50
#define X1_ADDON_PWR_EN_GPIO	IMX_GPIO_NR(2, 4)
#define ADDON_VENDOR_ATMARK_TECHNO	0x0001
#define ADDON_PRODUCT_ATMARK_TECHNO_SD	0x000a

static int x1_addon_init(void)
{
	struct addon_eeprom {
		unsigned short vendor;
		unsigned short product;
	};

	int ret;
	unsigned int bus;
	struct addon_eeprom eeprom_data;

	imx_iomux_v3_setup_multiple_pads(x1_addon_pwr_en,
					 ARRAY_SIZE(x1_addon_pwr_en));
	gpio_direction_output(X1_ADDON_PWR_EN_GPIO, 1);

	bus = i2c_get_bus_num();

	i2c_force_idle_bus(&i2c_pad_x1_addon_info3);
	ret = setup_i2c(0, CONFIG_SYS_I2C_SPEED, 0x7f, &i2c_pad_x1_addon_info3);
	if (ret)
		return 0;

	i2c_set_bus_num(0);

	i2c_init(CONFIG_SYS_I2C_SPEED, X1_ADDON_EEPROM_SLAVE);

	if (!i2c_probe(X1_ADDON_EEPROM_SLAVE)) {
		printf("X1 Addon EEPROM Detect\n");
		if (i2c_read(X1_ADDON_EEPROM_SLAVE, 0, 0, (uchar *)&eeprom_data, sizeof(eeprom_data)))
			printf("Addon EEPROM read error!!\n");
		else if ((eeprom_data.vendor == ADDON_VENDOR_ATMARK_TECHNO) &&
		    (eeprom_data.product == ADDON_PRODUCT_ATMARK_TECHNO_SD)) {
			printf("Atmark Techno Ext SD Slot Detect\n");
			x1_board_info.exist_sdhc1 = 1;
		}
	}

	i2c_set_bus_num(bus);

	return 0;
}

#ifdef CONFIG_FSL_QSPI
static iomux_v3_cfg_t const quadspi_pads[] = {
	MX7D_PAD_EPDC_DATA00__QSPI_A_DATA0 | MUX_PAD_CTRL(QSPI_PAD_CTRL),
	MX7D_PAD_EPDC_DATA01__QSPI_A_DATA1 | MUX_PAD_CTRL(QSPI_PAD_CTRL),
	MX7D_PAD_EPDC_DATA02__QSPI_A_DATA2 | MUX_PAD_CTRL(QSPI_PAD_CTRL),
	MX7D_PAD_EPDC_DATA03__QSPI_A_DATA3 | MUX_PAD_CTRL(QSPI_PAD_CTRL),
	MX7D_PAD_EPDC_DATA05__QSPI_A_SCLK  | MUX_PAD_CTRL(QSPI_PAD_CTRL),
	MX7D_PAD_EPDC_DATA06__QSPI_A_SS0_B | MUX_PAD_CTRL(QSPI_PAD_CTRL),
};

int board_qspi_init(void)
{
	/* Set the iomux */
	imx_iomux_v3_setup_multiple_pads(quadspi_pads,
					 ARRAY_SIZE(quadspi_pads));

	/* Set the clock */
	set_clk_qspi();

	return 0;
}
#endif

int board_early_init_f(void)
{
	setup_iomux_uart();

	imx_iomux_v3_setup_multiple_pads(usb_otg1_pads,
					 ARRAY_SIZE(usb_otg1_pads));

	return 0;
}

int board_init(void)
{
	/* address of boot parameters */
	gd->bd->bi_boot_params = PHYS_SDRAM + 0x100;

	imx_iomux_v3_setup_multiple_pads(iox_pads, ARRAY_SIZE(iox_pads));

	iox74lv_init();

	show_boot_source();

	i2c_force_idle_bus(&i2c_pad_info4);
	setup_i2c(0, CONFIG_SYS_I2C_SPEED, 0x7f, &i2c_pad_info4);
	board_info_init();

	if (x1_board_info.board_type == BOARDTYPE_AX1)
		x1_addon_init();

#ifdef CONFIG_FEC_MXC
	if (x1_board_info.board_type == BOARDTYPE_AIOTG_G3L) {
		gpio_request(PHY_RST_N_G3L, "lan1_rst_n");
		gpio_direction_output(PHY_RST_N_G3L, 1);
	} else if ((x1_board_info.board_type == BOARDTYPE_AX1) ||
		   (x1_board_info.board_type == BOARDTYPE_AIOTG_G3) ||
		   (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_D2) ||
		   (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_W1) ||
		   (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_M1) ||
		   (x1_board_info.board_type == BOARDTYPE_AIOTG_G3_W2)) {
		gpio_request(fec2_vsc8501_gpios.nreset, "lan1_rst_n");
		gpio_direction_output(fec2_vsc8501_gpios.nreset, 1);
	}

#ifdef CONFIG_X1_2ND_ETHERNET
	gpio_request(fec1_vsc8501_gpios.nreset, "lan2_rst_n");
	gpio_direction_output(fec1_vsc8501_gpios.nreset, 1);
#endif

	mdelay(100);

	if ((x1_board_info.board_type != BOARDTYPE_UNKNOWN) &&
	    (x1_board_info.board_type != BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z))
		setup_fec();
#endif

#ifdef CONFIG_NAND_MXS
	setup_gpmi_nand();
#endif

#ifdef CONFIG_FSL_QSPI
	board_qspi_init();
#endif

	return 0;
}

#define AIOT_G3_LED3	IMX_GPIO_NR(3, 11)
#define AIOT_G3_LED2	IMX_GPIO_NR(3, 10)
#define AIOT_G3_LED4	IMX_GPIO_NR(3, 12)
#define AIOT_G3_LED5	IMX_GPIO_NR(3, 13)

static int aiot_g3_led_init(void)
{
	imx_iomux_v3_setup_multiple_pads(gpio_aiot_g3_led_pads,
					 ARRAY_SIZE(gpio_aiot_g3_led_pads));

	gpio_direction_output(AIOT_G3_LED3, 1);
	gpio_direction_output(AIOT_G3_LED2, 0);
	gpio_direction_output(AIOT_G3_LED4, 0);
	gpio_direction_output(AIOT_G3_LED5, 0);

	return 0;

}

#define AIOT_G3L_LED3	IMX_GPIO_NR(3, 10)
#define AIOT_G3L_LED4	IMX_GPIO_NR(3, 11)
#define AIOT_G3L_LED5	IMX_GPIO_NR(3, 12)

static int aiot_g3l_led_init(void)
{
	imx_iomux_v3_setup_multiple_pads(gpio_aiot_g3l_led_pads,
					 ARRAY_SIZE(gpio_aiot_g3l_led_pads));

	gpio_direction_output(AIOT_G3L_LED3, 0);
	gpio_direction_output(AIOT_G3L_LED4, 1);
	gpio_direction_output(AIOT_G3L_LED5, 0);

	return 0;

}

#define AX1_LED_GREEN	IMX_GPIO_NR(2, 7)

static int ax1_led_init(void)
{
	imx_iomux_v3_setup_multiple_pads(gpio_ax1_led_pads,
					 ARRAY_SIZE(gpio_ax1_led_pads));

	gpio_direction_output(AX1_LED_GREEN, 1);

	return 0;

}

static int ax1_pcie_init(void)
{
	imx_iomux_v3_setup_multiple_pads(ax1_pcie_pads,
					 ARRAY_SIZE(ax1_pcie_pads));

	gpio_direction_output(AX1_PCIE_CLKREQ, 0);
	gpio_direction_output(AX1_PCIE_WAKE, 0);

	return 0;
}

#ifdef CONFIG_PFUZE3000_PMIC_I2C
#define PFUZE_DEVICEID	0x0
#define PFUZE_REVID	0x3
#define PFUZE_FABID	0x4

#define PFUZE_LDOGCTL	0x69

#define USB_HUB_RESET_N		IMX_GPIO_NR(2, 13)
#define PMIC_I2C_SLAVE_DEFAULT	0x08
#define PFUZE_PAGE		0x7f
#define 	PFUZE_PAGE_FUNCTIONAL_PAGE	0x0
#define 	PFUZE_PAGE_EXTENDED_PAGE1	0x1
#define 	PFUZE_PAGE_EXTENDED_PAGE2	0x2
#define PFUZE_OTP_I2C_ADDR	0xff
#define 	PFUZE_OTP_I2C_ADDR_USE_DEFAULT	(1<<7)
#define 	PFUZE_OTP_I2C_ADDR_SLB_ADDR_MASK	(0x03)
void change_pmic_address(void)
{
	unsigned char value;

	/* assert USB Hub reset */
	imx_iomux_v3_setup_pad(MX7D_PAD_EPDC_DATA13__GPIO2_IO13);
	gpio_direction_output(USB_HUB_RESET_N, 0);

	i2c_set_bus_num(CONFIG_PMIC_I2C_BUS);

	i2c_init(CONFIG_SYS_I2C_SPEED, PMIC_I2C_SLAVE_DEFAULT);
	if (!i2c_probe(PMIC_I2C_SLAVE_DEFAULT)) {
		value = PFUZE_PAGE_EXTENDED_PAGE1;
		if (i2c_write(PMIC_I2C_SLAVE_DEFAULT, PFUZE_PAGE, 1, &value, 1)) {
			printf("Write Extended Page1 error!\n");
			return;
		}

		if (i2c_read(PMIC_I2C_SLAVE_DEFAULT, PFUZE_OTP_I2C_ADDR, 1, &value, 1)) {
			printf("Read I2C address error!\n");
			return;
		}
		if (value & PFUZE_OTP_I2C_ADDR_USE_DEFAULT)
			return;

		value |= (CONFIG_PMIC_I2C_SLAVE - PMIC_I2C_SLAVE_DEFAULT) &
			PFUZE_OTP_I2C_ADDR_SLB_ADDR_MASK;
		if (i2c_write(PMIC_I2C_SLAVE_DEFAULT, PFUZE_OTP_I2C_ADDR, 1, &value, 1)) {
			printf("Write I2C address error!\n");
			return;
		}

		value = PFUZE_PAGE_FUNCTIONAL_PAGE;
		if (i2c_write(CONFIG_PMIC_I2C_SLAVE, PFUZE_PAGE, 1, &value, 1)) {
			printf("Write Default Page error!\n");
			return;
		}
	}
}

static int setup_pmic_voltages(void)
{
	unsigned char value, rev_id = 0;

	change_pmic_address();

	i2c_set_bus_num(CONFIG_PMIC_I2C_BUS);

	i2c_init(CONFIG_SYS_I2C_SPEED, CONFIG_PMIC_I2C_SLAVE);
	if (!i2c_probe(CONFIG_PMIC_I2C_SLAVE)) {
		if (i2c_read(CONFIG_PMIC_I2C_SLAVE, PFUZE_DEVICEID, 1, &value, 1)) {
			printf("Read device ID error!\n");
			return -1;
		}
		if (i2c_read(CONFIG_PMIC_I2C_SLAVE, PFUZE_REVID, 1, &rev_id, 1)) {
			printf("Read Rev ID error!\n");
			return -1;
		}
		printf("Found PFUZE300! deviceid 0x%x, revid 0x%x\n", value, rev_id);

		/* disable Low Power Mode during standby mode */
		if (i2c_read(CONFIG_PMIC_I2C_SLAVE, PFUZE_LDOGCTL, 1, &value, 1)) {
			printf("Read LDOCTL error!\n");
			return -1;
		}
		value |= 0x1;
		if (i2c_write(CONFIG_PMIC_I2C_SLAVE, PFUZE_LDOGCTL, 1, &value, 1)) {
			printf("Set LDOCTL error!\n");
			return -1;
		}
	}

	return 0;
}
#endif

#define USBNC_UH_HSICPHY_CFG1	0x30B30240
static void usb3503_detect_workaround(void)
{
	uint32_t tmp;
	tmp = readl(USBNC_UH_HSICPHY_CFG1);
	tmp &= ~0x00000f00;
	writel(tmp, USBNC_UH_HSICPHY_CFG1);
}

static void set_fdt_file(void)
{
	if (env_get("fdt_file"))
		return;

	switch (x1_board_info.board_type) {
	case BOARDTYPE_AX1:
		env_set("fdt_file", "armadillo_x1.dtb");
		break;
	case BOARDTYPE_AIOTG_G3L:
		env_set("fdt_file", "armadillo_iotg_g3l.dtb");
		break;
	case BOARDTYPE_AIOTG_G3:
		env_set("fdt_file", "armadillo_iotg_g3.dtb");
		break;
	case BOARDTYPE_AIOTG_G3_D2:
		env_set("fdt_file", "armadillo_iotg_g3_d2.dtb");
		break;
	case BOARDTYPE_AIOTG_G3_W1:
		env_set("fdt_file", "armadillo_iotg_g3_w1.dtb");
		break;
	case BOARDTYPE_AIOTG_G3_M1:
		env_set("fdt_file", "armadillo_iotg_g3_m1.dtb");
		break;
	case BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z:
		env_set("fdt_file", "armadillo_iotg_g3_d2_agx3031_e00z.dtb");
		break;
	case BOARDTYPE_AIOTG_G3_W2:
		env_set("fdt_file", "armadillo_iotg_g3_w2.dtb");
		break;
	default:
		env_set("fdt_file", "x1-minimal.dtb");
		break;
	}
}

static void set_mmcdev(void)
{
	enum boot_device boot_dev;

	if (env_get("mmcdev"))
		return ;

	if (!x1_board_info.exist_sdhc1) {
		env_set("mmcdev", "0");
		return ;
	}

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case SD1_BOOT:
		/* USDHC1 */
		env_set("mmcdev", "0");
		return ;
	default:
		/* eMMC (Becomes "0" when SDHC1 does not exist)*/
		env_set("mmcdev", "1");
		return ;
	}

}

static void set_mmcroot(void)
{
	enum boot_device boot_dev;

	if (env_get("mmcroot"))
		return ;

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case SD1_BOOT:
		env_set("mmcroot", "/dev/mmcblk0p2 rootwait rw");
		break;
	default:
		env_set("mmcroot", "/dev/mmcblk2p2 rootwait rw");
		break;
	}
}

static void set_board_info(void)
{
	char revision[sizeof(x1_board_info.revision)];
	char serial_no[sizeof(x1_board_info.serial_no)];
	char kotai[12];

	sprintf(revision, "0x%04x", x1_board_info.revision);
	env_set("revision#", revision);
	if (x1_board_info.serial_no_valid == SVALID_SERIALNO) {
		sprintf(serial_no, "%d", x1_board_info.serial_no);
		env_set("serial#", serial_no);
	}
	else if (x1_board_info.serial_no_valid == SVALID_KOTAI) {
		sprintf(kotai, "%04X%04d%04d", x1_board_info.kotai.product_id,
				x1_board_info.kotai.lot, x1_board_info.kotai.number);
		env_set("serial#", kotai);
	}
}

int board_late_init(void)
{
#ifdef CONFIG_PFUZE3000_PMIC_I2C
	int ret = 1;

	ret = setup_pmic_voltages();
	if (ret)
		return ret;
#endif

	imx_iomux_v3_setup_multiple_pads(wdog_pads, ARRAY_SIZE(wdog_pads));
	set_wdog_reset((struct wdog_regs *)WDOG1_BASE_ADDR);
	hw_watchdog_init();

	imx_iomux_v3_setup_multiple_pads(maintenance_mode_gpio_pads,
					 ARRAY_SIZE(maintenance_mode_gpio_pads));
	gpio_direction_input(MAINTENANCE_MODE_GPIO);

	switch (x1_board_info.board_type) {
	case BOARDTYPE_AX1:
		ax1_led_init();
		ax1_pcie_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3L:
		aiot_g3l_led_init();
		break;
	case BOARDTYPE_AIOTG_G3:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3_D2:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3_W1:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3_M1:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3_D2_AGX3031_E00Z:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	case BOARDTYPE_AIOTG_G3_W2:
		aiot_g3_led_init();
		usb3503_detect_workaround();
		break;
	default:
		/* do nothing */
		break;
	}

	set_fdt_file();
	set_mmcdev();
	set_mmcroot();
	set_board_info();

	return 0;
}

#ifdef CONFIG_USB_EHCI_MX7
int board_usb_phy_mode(int port)
{
	return USB_INIT_HOST;
}

int board_ehci_power(int port,int on)
{
	return gpio_direction_output(G_OTG1_PWREN, on);
}
#endif

#ifdef CONFIG_SUPPORT_MAINTENANCE_MODE
int is_maintenance_mode(void)
{
	return !gpio_get_value(MAINTENANCE_MODE_GPIO);
}
#endif /* CONFIG_SUPPORT_MAINTENANCE_MODE */

#ifdef CONFIG_SPL_BUILD
#include <libfdt.h>
#include <spl.h>

u32 spl_boot_device(void)
{
	enum boot_device boot_dev;

	boot_dev = get_boot_device();
	switch (boot_dev) {
	case SD1_BOOT:
		return BOOT_DEVICE_MMC1;
	case MMC3_BOOT:
		return BOOT_DEVICE_MMC2;
	case QSPI_BOOT:
		/* Manufacture boot - see IMX7DRM Figure 6-47 */
		return BOOT_DEVICE_SPI;
	default:
		return BOOT_DEVICE_NONE;
	}
}

/* MT41K128M16JT-125 IT:K */
static void spl_dram_init_ddr3_32bit_512mbytes(void)
{
	/* assert DDR controller reset */
	writel(0x00000003, 0x30391000);
	/* de-assert dfi_init_start and ctrl_wake_up[4:0] */
	writel(0x00000000, 0x30340020);

	/* Clear then set bit30 to ensure exit from DDR retention */
	writel(0x40000000, 0x30360388);
	writel(0x40000000, 0x30360384);

	writel(0x00000002, 0x30391000);
	writel(0x01040001, 0x307a0000);
	writel(0x80400003, 0x307a01a0);
	writel(0x00100020, 0x307a01a4);
	writel(0x80100004, 0x307a01a8);

	/*
	 * T_RFC_NOM_X32 = RoundDown(tREFI / tCK / 32 / 2 )
	 *               = RoundDown(7.8us / 1.876173 / 32 / 2)
	 *               = 64 = 0x40
	 *
	 * T_RFC_MIN = RoundUp(RoundUp(tRFCmin / tCK) / 2)
	 *           = RoundUp (RoundUp (160ns / 1.876173) / 2)
	 *           = RoundUp (86 / 2)
	 *           = 43 = 0x2b
	 *
	 * tCK = 1 / 533MHz = 1.876173(ns)
	 */
	writel(0x0040002b, 0x307a0064);
	writel(0x00000001, 0x307a0490);
	writel(0x00020083, 0x307a00d0);
	writel(0x00690000, 0x307a00d4);
	writel(0x09300004, 0x307a00dc);
	writel(0x04080000, 0x307a00e0);
	writel(0x00100004, 0x307a00e4);
	writel(0x0000033f, 0x307a00f4);
	writel(0x0908110a, 0x307a0100);
	writel(0x0007020e, 0x307a0104);
	writel(0x03040407, 0x307a0108);
	writel(0x00002006, 0x307a010c);
	writel(0x04020205, 0x307a0110);
	writel(0x03030202, 0x307a0114);
	writel(0x00000803, 0x307a0120);
	writel(0x00800020, 0x307a0180);
	writel(0x02098204, 0x307a0190);
	writel(0x00030303, 0x307a0194);
	/*
	 * ADDRMAP_CS_BIT1: 21(0x15)
	 * Internal Base  : 6
	 * HIF address bit: 27(21 + 6)
	 */
	writel(0x00000015, 0x307a0200);
	/*
	 * BANK_B2: bit26
	 * BANK_B1: bit25
	 * BANK_B0: bit24
	 */
	writel(0x00161616, 0x307a0204);
	/*
	 * ROW_B11: bit21
	 * ROW_B10: bit20
	 *     :
	 * ROW_B2 : bit12
	 * ROW_B1 : bit11
	 * ROW_B0 : bit10
	 */
	writel(0x04040404, 0x307a0214);
	/*
	 * ROW_B13: bit23
	 * ROW_B12: bit22
	 */
	writel(0x00000404, 0x307a0218);
	writel(0x06000604, 0x307a0240);
	writel(0x00000001, 0x307a0244);
	writel(0x00000000, 0x30391000);
	writel(0x17420f40, 0x30790000);
	writel(0x10210100, 0x30790004);
	writel(0x00060807, 0x30790010);
	writel(0x1010007e, 0x307900b0);
	writel(0x000006db, 0x3079009c);
	writel(0x08080808, 0x30790020);
	writel(0x08080808, 0x30790030);
	writel(0x01000010, 0x30790050);
	writel(0x00000010, 0x30790050);

	writel(0x0e407304, 0x307900c0);
	writel(0x0e447304, 0x307900c0);
	writel(0x0e447306, 0x307900c0);

	if (wait_for_bit(__func__, (u32 *)0x307900c4, 0x01, true, 10000, 0)) {
		printf("timeout waiting for DDR3 ZQ Calibration is finished.\n");
		reset_cpu(0);
	}

	writel(0x0e447304, 0x307900c0);
	writel(0x0e407304, 0x307900c0);

	writel(0x00000000, 0x30384130);
	writel(0x00000178, 0x30340020);
	writel(0x00000002, 0x30384130);
	writel(0x0000000f, 0x30790018);

	if (wait_for_bit(__func__, (u32 *)0x307a0004, 0x01, true, 10000, 0)) {
		printf("timeout waiting for DDR3 operating mode becomes normal.\n");
		reset_cpu(0);
	}
}

/* MT41K256M16TW-107 IT:P */
static void spl_dram_init_ddr3_32bit_1gbytes(void)
{
	/* assert DDR controller reset */
	writel(0x00000003, 0x30391000);
	/* de-assert dfi_init_start and ctrl_wake_up[4:0] */
	writel(0x00000000, 0x30340020);

	/* Clear then set bit30 to ensure exit from DDR retention */
	writel(0x40000000, 0x30360388);
	writel(0x40000000, 0x30360384);

	writel(0x00000002, 0x30391000);
	writel(0x01040001, 0x307a0000);
	writel(0x80400003, 0x307a01a0);
	writel(0x00100020, 0x307a01a4);
	writel(0x80100004, 0x307a01a8);

	/*
	 * T_RFC_NOM_X32 = RoundDown(tREFI / tCK / 32 / 2 )
	 *               = RoundDown(7.8us / 1.876173 / 32 / 2)
	 *               = 64 = 0x40
	 *
	 * T_RFC_MIN = RoundUp(RoundUp(tRFCmin / tCK) / 2)
	 *           = RoundUp (RoundUp (260ns / 1.876173) / 2)
	 *           = RoundUp (139 / 2)
	 *           = 70 = 0x46
	 *
	 * tCK = 1 / 533MHz = 1.876173(ns)
	 */
	writel(0x00400046, 0x307a0064);
	writel(0x00000001, 0x307a0490);
	writel(0x00020083, 0x307a00d0);
	writel(0x00690000, 0x307a00d4);
	writel(0x09300004, 0x307a00dc);
	writel(0x04080000, 0x307a00e0);
	writel(0x00100004, 0x307a00e4);
	writel(0x0000033f, 0x307a00f4);
	writel(0x0908110a, 0x307a0100);
	writel(0x0007020e, 0x307a0104);
	writel(0x03040407, 0x307a0108);
	writel(0x00002006, 0x307a010c);
	writel(0x04020205, 0x307a0110);
	writel(0x03030202, 0x307a0114);
	writel(0x00000803, 0x307a0120);
	writel(0x00800020, 0x307a0180);
	writel(0x02098204, 0x307a0190);
	writel(0x00030303, 0x307a0194);
	/*
	 * ADDRMAP_CS_BIT1: 22(0x16)
	 * Internal Base  : 6
	 * HIF address bit: 28(22 + 6)
	 */
	writel(0x00000016, 0x307a0200);
	/*
	 * BANK_B2: bit27
	 * BANK_B1: bit26
	 * BANK_B0: bit25
	 */
	writel(0x00171717, 0x307a0204);
	/*
	 * ROW_B11: bit21
	 * ROW_B10: bit20
	 *     :
	 * ROW_B2 : bit12
	 * ROW_B1 : bit11
	 * ROW_B0 : bit10
	 */
	writel(0x04040404, 0x307a0214);
	/*
	 * ROW_B14: bit24
	 * ROW_B13: bit23
	 * ROW_B12: bit22
	 */
	writel(0x00040404, 0x307a0218);
	writel(0x06000604, 0x307a0240);
	writel(0x00000001, 0x307a0244);
	writel(0x00000000, 0x30391000);
	writel(0x17420f40, 0x30790000);
	writel(0x10210100, 0x30790004);
	writel(0x00060807, 0x30790010);
	writel(0x1010007e, 0x307900b0);
	writel(0x000006db, 0x3079009c);
	writel(0x08080808, 0x30790020);
	writel(0x08080808, 0x30790030);
	writel(0x01000010, 0x30790050);
	writel(0x00000010, 0x30790050);

	writel(0x0e407304, 0x307900c0);
	writel(0x0e447304, 0x307900c0);
	writel(0x0e447306, 0x307900c0);

	if (wait_for_bit(__func__, (u32 *)0x307900c4, 0x01, true, 10000, 0)) {
		printf("timeout waiting for DDR3 ZQ Calibration is finished.\n");
		reset_cpu(0);
	}

	writel(0x0e447304, 0x307900c0);
	writel(0x0e407304, 0x307900c0);

	writel(0x00000000, 0x30384130);
	writel(0x00000178, 0x30340020);
	writel(0x00000002, 0x30384130);
	writel(0x0000000f, 0x30790018);

	if (wait_for_bit(__func__, (u32 *)0x307a0004, 0x01, true, 10000, 0)) {
		printf("timeout waiting for DDR3 operating mode becomes normal.\n");
		reset_cpu(0);
	}
}

static void spl_dynamic_dram_init(void)
{
	long size;

	spl_dram_init_ddr3_32bit_1gbytes();
	size = get_ram_size((void *)PHYS_SDRAM, SZ_1G);
	if (size == SZ_1G)
		return;

	spl_dram_init_ddr3_32bit_512mbytes();
	size = get_ram_size((void *)PHYS_SDRAM, SZ_512M);
	if (size == SZ_512M)
		return;

	printf("unsupported dram setting\n");
	hang();
}

#define DDR_TYPE_DDR3	0x00
#define DDR_TYPE_LPDDR2	0x01
static int spl_dram_init(unsigned int dram)
{
	unsigned int width = (1 << (dram & 0xff));
	unsigned int size = (1 << ((dram & 0xff00) >> 8));
	unsigned int type = ((dram & 0xff0000) >> 16);

	if (!dram) {
		spl_dynamic_dram_init();
		return 0;
	}

	if (width != 32) {
		printf("unsupported dram bus width %d\n", width);
		hang();
	}

	if (type != DDR_TYPE_DDR3) {
		printf("unsupported dram type %02x\n", type);
		hang();
	}

	switch (size) {
	case SZ_512M:
		spl_dram_init_ddr3_32bit_512mbytes();
		break;
	case SZ_1G:
		spl_dram_init_ddr3_32bit_1gbytes();
		break;
	default:
		printf("unsupported dram size %d\n", size);
		hang();
	}

	return 0;
}

void board_init_f(ulong dummy)
{
	struct iomuxc_gpr_base_regs *const iomuxc_gpr_regs
		= (struct iomuxc_gpr_base_regs *) IOMUXC_GPR_BASE_ADDR;

	arch_cpu_init();

	board_early_init_f();

	/* UART clocks enabled and gd valid - init serial console */
	preloader_console_init();

	/* Enable On-chip RAM EPDC Function */
	setbits_le32(&iomuxc_gpr_regs->gpr[1],
		     IOMUXC_GPR_GPR1_GPR_ENABLE_OCRAM_EPDC_MASK);

	/* Clear the BSS. */
	memset(__bss_start, 0, __bss_end - __bss_start);

	/* load/boot image from boot device */
	board_init_r(NULL, 0);
}

void spl_board_init(void)
{
	unsigned int dram = 0;

	i2c_force_idle_bus(&i2c_pad_info4);
	setup_i2c(0, CONFIG_SYS_I2C_SPEED, 0x7f, &i2c_pad_info4);
	bmic_init();
	if (!bmic_eeprom_init())
		bmic_eeprom_read(16, &dram, sizeof(dram));

	set_clk_a7(996000);

	/* DDR initialization */
	spl_dram_init(dram);
}

#endif /* CONFIG_SPL_BUILD */
